use futures::future;
use rust_bfd::events::{EventMessageSink, ScriptHookSink, StateChangeEvent};
use rust_bfd::{set_ttl_or_hop_limit, BfdSessionStats};
use rust_bfd::{bfd::*, events::EventMessage};
use std::net::Ipv6Addr;
use std::{
    collections::HashMap,
    error::Error,
    fmt::Display,
    io::ErrorKind,
    mem::swap,
    net::{IpAddr, SocketAddr},
    str::FromStr,
    sync::{atomic::Ordering, Arc},
    time::{Duration, Instant},
};

// Necessary until a hop_limit method is added to UdpSocket
use nix::sys::socket::{setsockopt, sockopt};

use tokio::sync::{oneshot, Mutex};

use env_logger::Env;
use log::{debug, error, info, warn};

use nom_derive::Parse;
use rand::prelude::*;
use tokio::io;
use tokio::net::UdpSocket;
use tokio::sync::mpsc;
use tokio::task::{self, JoinHandle};
use tokio::time;

use atomic::Atomic;

const CONTROL_PORT: u16 = 3784;
const ECHO_PORT: u16 = 3785;

const ORDERING: Ordering = Ordering::Relaxed;

#[derive(Debug)]
pub enum BfdError {
    // field, value
    InvalidFieldValue(&'static str, &'static str),
}
impl Display for BfdError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::InvalidFieldValue(field, value) => {
                write!(f, "invalid value `{}` for field `{}`", field, value)
            }
        }
    }
}
impl Error for BfdError {}

/// Data structure to store the state of the Bfd machine. The impl does *not*
/// implement any Bfd logic, it is merely a thread-safe data structure
#[derive(Debug)]
struct BfdSessionState {
    control_sock: Arc<UdpSocket>,                  // immutable
    peer_addr: IpAddr,                             // immutable and Copy
    session_state: Atomic<BfdState>,               // mutable by any thread
    remote_session_state: Atomic<BfdState>,        // mutable by rx thread
    local_discr: BfdDiscriminator,                 // immutable and Copy
    remote_discr: Atomic<BfdDiscriminator>,        // mutable by rx thread
    local_diag: Atomic<BfdDiagnostic>,             // mutable by any thread
    desired_min_tx_interval: Atomic<BfdInterval>,  // mutable by any thread
    required_min_rx_interval: Atomic<BfdInterval>, // mutable by any thread
    remote_min_rx_interval: Atomic<BfdInterval>,   // mutable by rx thread
    demand_mode: Atomic<bool>,                     // mutable by rx thread
    remote_demand_mode: Atomic<bool>,              // mutable by rx thread
    detect_mult: Atomic<u8>,                       // mutable by control thread
    auth_type: Atomic<BfdAuthType>,                // mutable by control thread
    rcv_auth_seq: Atomic<u32>,                     // mutable by rx thread
    xmit_auth_seq: Atomic<u32>,                    // mutable by tx thread?
    auth_seq_known: Atomic<bool>,                  // mutable by rx thread?

    detection_time: Atomic<BfdInterval>, // mutable by rx thread
    poll_mode: Atomic<bool>,             // mutable by rx thread

    control_packets_rx: Atomic<u64>,
    control_packets_tx: Atomic<u64>,
    last_state_change: Mutex<Instant>,
    last_remote_diag: Atomic<BfdDiagnostic>,
}

// TODO: hide the internal data structure behind a trait
impl BfdSessionState {
    fn control_sock(&self) -> Arc<UdpSocket> {
        self.control_sock.clone()
    }
    fn peer_addr(&self) -> IpAddr {
        self.peer_addr
    }

    fn session_state(&self) -> BfdState {
        self.session_state.load(ORDERING)
    }
    // Note: This does not handle state transition
    fn set_session_state(&self, value: BfdState) {
        self.session_state.store(value, ORDERING)
    }
    fn remote_session_state(&self) -> BfdState {
        self.remote_session_state.load(ORDERING)
    }
    fn set_remote_session_state(&self, value: BfdState) {
        self.remote_session_state.store(value, ORDERING);
    }
    fn local_discr(&self) -> BfdDiscriminator {
        self.local_discr
    }
    fn remote_discr(&self) -> BfdDiscriminator {
        self.remote_discr.load(ORDERING)
    }
    fn set_remote_discr(&self, value: BfdDiscriminator) {
        self.remote_discr.store(value, ORDERING);
    }
    fn demand_mode(&self) -> bool {
        self.demand_mode.load(ORDERING)
    }
    fn set_demand_mode(&self, value: bool) {
        self.demand_mode.store(value, ORDERING)
    }
    fn remote_demand_mode(&self) -> bool {
        self.remote_demand_mode.load(ORDERING)
    }
    fn set_remote_demand_mode(&self, value: bool) {
        self.remote_demand_mode.store(value, ORDERING);
    }
    fn remote_min_rx_interval(&self) -> BfdInterval {
        self.remote_min_rx_interval.load(ORDERING)
    }
    fn set_remote_min_rx_interval(&self, value: BfdInterval) {
        self.remote_min_rx_interval.store(value, ORDERING);
    }
    fn local_diag(&self) -> BfdDiagnostic {
        self.local_diag.load(ORDERING)
    }
    fn set_local_diag(&self, value: BfdDiagnostic) {
        self.local_diag.store(value, ORDERING);
    }
    fn auth_type(&self) -> BfdAuthType {
        self.auth_type.load(ORDERING)
    }
    fn detect_mult(&self) -> u8 {
        self.detect_mult.load(ORDERING)
    }
    fn set_detect_mult(&self, value: u8) {
        self.detect_mult.store(value, ORDERING)
    }
    fn poll_mode(&self) -> bool {
        self.poll_mode.load(ORDERING)
    }
    fn set_poll_mode(&self, value: bool) {
        self.poll_mode.store(value, ORDERING)
    }
    fn desired_min_tx_interval(&self) -> BfdInterval {
        self.desired_min_tx_interval.load(ORDERING)
    }
    fn set_desired_min_tx_interval(&self, value: BfdInterval) {
        self.desired_min_tx_interval.store(value, ORDERING);
    }
    fn required_min_rx_interval(&self) -> BfdInterval {
        self.required_min_rx_interval.load(ORDERING)
    }
    fn set_required_min_rx_interval(&self, value: BfdInterval) {
        self.required_min_rx_interval.store(value, ORDERING)
    }
    fn detection_time(&self) -> BfdInterval {
        self.detection_time.load(ORDERING)
    }
    fn set_detection_time(&self, value: BfdInterval) {
        self.detection_time.store(value, ORDERING)
    }
    fn control_packets_rx(&self) -> u64 {
        self.control_packets_rx.load(ORDERING)
    }
    fn control_packets_tx(&self) -> u64 {
        self.control_packets_tx.load(ORDERING)
    }
    fn last_state_change(&self) -> Instant {
        *self.last_state_change.try_lock().unwrap()
    }
    fn last_remote_diag(&self) -> BfdDiagnostic {
        self.last_remote_diag.load(ORDERING)
    }
    fn set_last_remote_diag(&self, value: BfdDiagnostic) {
        self.last_remote_diag.store(value, ORDERING)
    }
}

enum WatchdogReset {
    ReceivedPacket,
    Terminate,
}

struct BfdSession {
    state: BfdSessionState,
    periodic: BfdPeriodicSenderHandle,
    rx: mpsc::Receiver<SessionControlCommand>,
    tx: mpsc::Sender<SessionControlCommand>,
    rx_watchdog: Option<mpsc::Sender<WatchdogReset>>,
    cur_interval: BfdInterval,
    event_dispatcher: EventDispatcherHandle,
}

impl From<&BfdSessionState> for BfdSessionStats {
    fn from(state: &BfdSessionState) -> Self {
        Self {
            local_ip: state.control_sock().local_addr().unwrap().ip(),
            remote_ip: state.peer_addr(),
            local_discr: state.local_discr(),
            remote_discr: state.remote_discr(),
            state: state.session_state(),
            last_diag: state.last_remote_diag(),
            control_packets_rx: state.control_packets_rx(),
            control_packets_tx: state.control_packets_tx(),
            last_change: state.last_state_change(),
            detect_time: state.detection_time(),
            base_interval: BfdInterval::from_secs(0), // TODO: This is a hack and will be patched up in the session
        }
    }
}
enum PeriodicControlCommand {
    Stop,
    Start,
    Quit,
    SetMinInterval(BfdInterval),
    SendPollResponse,
}

enum SessionControlCommand {
    RxPacket(BfdPacket),
    TxControlPacket {
        poll_response: bool,
    },
    GetSessionStats {
        respond_to: oneshot::Sender<BfdSessionStats>,
    },
    ReceiveTimeout,
    Quit,
}

#[derive(Clone)]
struct BfdSessionHandle {
    tx: mpsc::Sender<SessionControlCommand>,
}

impl BfdSessionHandle {
    async fn new(
        local_addr: IpAddr,
        remote_addr: IpAddr,
        local_discr: BfdDiscriminator,
        event_dispatcher: EventDispatcherHandle,
    ) -> Self {
        let (tx, rx) = mpsc::channel(32);
        let new_self = Self { tx };
        let periodic =
            BfdPeriodicSenderHandle::new(new_self.clone(), time::Duration::from_secs(1).into());

        let mut session = BfdSession::new(
            rx,
            new_self.tx.clone(),
            periodic,
            local_addr,
            remote_addr,
            local_discr,
            event_dispatcher,
        )
        .await;
        tokio::spawn(async move { session.run().await });

        new_self
    }
    async fn transmit_control_packet(
        &self,
        poll_response: bool,
    ) -> Result<(), mpsc::error::SendError<SessionControlCommand>> {
        self.tx
            .send(SessionControlCommand::TxControlPacket { poll_response })
            .await
    }
    async fn get_stats(
        &self,
    ) -> Result<BfdSessionStats, mpsc::error::SendError<SessionControlCommand>> {
        let (send, recv) = oneshot::channel();
        self.tx
            .send(SessionControlCommand::GetSessionStats { respond_to: send })
            .await?;
        Ok(recv.await.expect("Unable to receive stats from session"))
    }
}

#[derive(Clone, Debug)]
struct BfdPeriodicSenderHandle {
    tx: mpsc::Sender<PeriodicControlCommand>,
}

impl BfdPeriodicSenderHandle {
    fn new(bfd: BfdSessionHandle, interval: BfdInterval) -> Self {
        let (tx, rx) = mpsc::channel(32);
        let mut periodic_sender = BfdPeriodicSender::new(rx, bfd, interval);
        tokio::spawn(async move { periodic_sender.run().await });

        Self { tx }
    }
}

struct BfdPeriodicSender {
    rx: mpsc::Receiver<PeriodicControlCommand>,
    bfd: BfdSessionHandle,
    cur_interval: BfdInterval,
}

impl BfdPeriodicSender {
    fn new(
        rx: mpsc::Receiver<PeriodicControlCommand>,
        bfd: BfdSessionHandle,
        interval: BfdInterval,
    ) -> Self {
        Self {
            rx,
            bfd,
            cur_interval: interval,
        }
    }
    async fn run(&mut self) {
        let base_interval = self.cur_interval * 3 / 4;
        let mut clock = time::interval(base_interval.into());
        let mut running = true;

        loop {
            tokio::select! {
             _ = clock.tick(), if running == true => {
                        let jitter = time::Duration::from_micros(
                            rand::thread_rng()
                                .gen_range(0..clock.period().as_micros() / 3)
                                .try_into()
                                .unwrap(),
                        );
                        time::sleep(jitter).await;
                        // debug!("tick! sent periodic control packet after {}ms", (base_interval + jitter).as_millis());
                        self.bfd.transmit_control_packet(false).await.unwrap();
                    },
                    Some(cmd) = self.rx.recv() => {
                        match cmd {
                            PeriodicControlCommand::Quit => { debug!("Closing periodic packet sender"); return},
                            PeriodicControlCommand::Stop => {
                                debug!("Stopping periodic packets");
                                running= false}
                            ,
                            PeriodicControlCommand::Start => { debug!("Starting periodic packets"); running = true },
                            PeriodicControlCommand::SetMinInterval(i) => {
                                let base_interval: Duration = (i * 3 / 4).into();
                                debug!("Updating base interval to {}ms (jittering {} to {}ms)", base_interval.as_millis(), base_interval.as_millis(), base_interval.as_millis() + base_interval.as_millis() / 3);
                                clock = time::interval_at(
                                    time::Instant::now() + base_interval.into(),
                                    base_interval.into(),
                                );
                            },
                            PeriodicControlCommand::SendPollResponse => {
                                debug!("Sending poll response");
                                self.bfd.transmit_control_packet(true).await.unwrap();
                            }
                        }
                    }
            };
        }
    }
}

impl BfdSession {
    async fn new(
        rx: mpsc::Receiver<SessionControlCommand>,
        tx: mpsc::Sender<SessionControlCommand>,
        periodic: BfdPeriodicSenderHandle,
        local_addr: IpAddr,
        remote_addr: IpAddr,
        local_discr: BfdDiscriminator,
        event_dispatcher: EventDispatcherHandle,
    ) -> Self {
        let mut rng = rand::thread_rng();

        //TODO: select a random unused port instead of pure random
        let source_port: u16 = rng.gen_range(49152..=65535);
        let control_sock = UdpSocket::bind(SocketAddr::new(local_addr, source_port))
            .await
            .unwrap();
        set_ttl_or_hop_limit(&control_sock, 255).unwrap();
        // control_sock
        //     .connect(SocketAddr::new(remote_addr, CONTROL_PORT))
        //     .await?;

        // Incoming packets will come over the channel from the mux, since they don't send to the reciprocal port
        Self {
            rx,
            tx,
            periodic,
            rx_watchdog: None,
            cur_interval: BfdInterval::from_secs(1),
            event_dispatcher,
            state: BfdSessionState {
                control_sock: Arc::new(control_sock),
                peer_addr: remote_addr,
                session_state: Atomic::new(BfdState::default()),
                remote_session_state: Atomic::new(BfdState::default()),
                local_discr,
                remote_discr: Atomic::new(BfdDiscriminator(0)),
                local_diag: Atomic::new(BfdDiagnostic::None),
                desired_min_tx_interval: Atomic::new(BfdInterval::from_millis(300)),
                required_min_rx_interval: Atomic::new(BfdInterval::from_millis(300)),
                remote_min_rx_interval: Atomic::new(BfdInterval::from_millis(300)),
                demand_mode: Atomic::new(false),
                remote_demand_mode: Atomic::new(false),
                detect_mult: Atomic::new(3),
                auth_type: Atomic::new(BfdAuthType::None),
                rcv_auth_seq: Atomic::new(0),
                xmit_auth_seq: Atomic::new(rng.gen()),
                auth_seq_known: Atomic::new(false),
                detection_time: Atomic::new(BfdInterval::from_secs(0)),
                poll_mode: Atomic::new(false),

                control_packets_rx: Atomic::new(0),
                control_packets_tx: Atomic::new(0),
                last_state_change: Mutex::new(Instant::now()),
                last_remote_diag: Atomic::new(BfdDiagnostic::None),
            },
        }
    }

    async fn transition(&mut self, new_state: BfdState) {
        if self.state.session_state() == new_state {
            return;
        };
        self.event_dispatcher
            .dispatch(EventMessage::StateChange(StateChangeEvent {
                local_discr: self.state.local_discr(),
                remote_discr: self.state.remote_discr(),
                local_ip: self.state.control_sock().local_addr().unwrap().ip(), //TODO: shouldn't unwrap here, it will crash the program
                remote_ip: self.state.peer_addr(),
                from_state: self.state.session_state(),
                to_state: new_state,
            }))
            .await;
        info!(
            "Peer {} state change {} -> {}",
            self.state.peer_addr(),
            self.state.session_state(),
            new_state
        );
        *self.state.last_state_change.lock().await = Instant::now();
        match (self.state.session_state(), new_state) {
            (BfdState::Up, BfdState::AdminDown | BfdState::Down | BfdState::Init) => {
                if self.rx_watchdog.is_some() {
                    self.rx_watchdog
                        .as_mut()
                        .unwrap()
                        .send(WatchdogReset::Terminate)
                        .await
                        .unwrap();
                };
                //   When bfd.SessionState is not Up, the system MUST set bfd.DesiredMinTxInterval to a value of not
                //    less than one second (1,000,000 microseconds).  This is intended to ensure that the bandwidth
                //    consumed by BFD sessions that are not Up is negligible, particularly in the case where a neighbor
                //    may not be running BFD.
                self.state.set_desired_min_tx_interval(std::cmp::min(
                    self.state.desired_min_tx_interval(),
                    BfdInterval::from_secs(1),
                ));
                self.update_transmit_interval().await;
            }
            (BfdState::AdminDown | BfdState::Down | BfdState::Init, BfdState::Up) => {
                self.start_watchdog().await;
                self.state.set_local_diag(BfdDiagnostic::None); // This doesn't seem to be required by the spec but seems sensible and at least FRR does it.
            }
            _ => {}
        }
        self.state.set_session_state(new_state);
    }

    async fn run(&mut self) {
        while let Some(cmd) = self.rx.recv().await {
            match cmd {
                SessionControlCommand::Quit => return,
                SessionControlCommand::RxPacket(packet) => {
                    self.receive_control_packet(&packet).await;
                }
                SessionControlCommand::GetSessionStats { respond_to } => {
                    let mut stats =
                        BfdSessionStats::from(&self.state).with_base_interval(self.cur_interval);
                    respond_to.send(stats).unwrap()
                }
                SessionControlCommand::TxControlPacket { poll_response } => {
                    self.transmit_control_packet(poll_response).await
                }
                SessionControlCommand::ReceiveTimeout => {
                    self.state.set_local_diag(BfdDiagnostic::TimeExpired);
                    self.transition(BfdState::Down).await;
                }
            }
        }
        println!("Session command channel closed");
    }

    async fn transmit_control_packet(&self, poll_response: bool) {
        let packet = BfdPacket {
            flags: BfdFlags(0)
                .with_vers(1)
                .with_diag(self.state.local_diag().into())
                .with_state(self.state.session_state().into())
                .with_poll(if poll_response {
                    false
                } else {
                    self.state.poll_mode()
                })
                .with_final_(poll_response)
                .with_cpi(true)
                .with_demand(
                    self.state.session_state() == BfdState::Up
                        && self.state.remote_session_state() == BfdState::Up,
                )
                .with_detect_mult(self.state.detect_mult())
                .with_length(24),
            my_disc: self.state.local_discr(),
            your_disc: self.state.remote_discr(),
            desired_min_tx: self.state.desired_min_tx_interval(),
            required_min_rx: self.state.required_min_rx_interval(),
            required_min_echo_rx: BfdInterval::from_secs(0),
            auth: None,
        };
        let socket = self.state.control_sock.clone();
        let dest = self.state.peer_addr;
        self.state.control_packets_tx.fetch_add(1, ORDERING);
        debug!("tx packet: {:?}", packet);
        match socket
            .send_to(
                packet.serialize().unwrap().as_ref(),
                SocketAddr::new(dest, CONTROL_PORT),
            )
            .await
        {
            Err(e) => {
                warn!(
                    "Error sending packet to {:?}:{} ({:?})",
                    dest, CONTROL_PORT, e
                );
            }
            Ok(_) => {}
        }
    }

    async fn start_watchdog(&mut self) {
        let (tx, mut rx) = mpsc::channel(1);
        let cmd_tx = self.tx.clone();
        let duration = self.state.detection_time().into();
        let peer_ip = self.state.peer_addr();

        tokio::spawn(async move {
            let mut running = true;
            while running {
                match tokio::time::timeout(duration, async { rx.recv().await }).await {
                    Ok(Some(WatchdogReset::ReceivedPacket)) => {
                        debug!("Packet received, watchdog ok")
                    }
                    Ok(Some(WatchdogReset::Terminate)) | Ok(None) => {
                        debug!("Watchdog terminating");
                        running = false
                    }
                    Err(_) => {
                        info!(
                            "Peer {} timed out after {}ms",
                            peer_ip,
                            duration.as_millis()
                        );
                        cmd_tx
                            .send(SessionControlCommand::ReceiveTimeout)
                            .await
                            .unwrap()
                    }
                }
            }
        });
        let mut temporary = Some(tx);
        swap(&mut temporary, &mut self.rx_watchdog);
        if temporary.is_some() {
            debug!(
                "Updating old watchdog with new detection time ({}ms)",
                duration.as_millis()
            );
            temporary
                .as_mut()
                .unwrap()
                .send(WatchdogReset::Terminate)
                .await
                .unwrap_or(())
        }
    }

    // https://datatracker.ietf.org/doc/html/rfc5880#section-6.8.6
    async fn receive_control_packet(&mut self, p: &BfdPacket) {
        // https://datatracker.ietf.org/doc/html/rfc5880#section-6.8.4
        //
        // If Demand mode is not active, and a period of time equal to the
        // Detection Time passes without receiving a BFD Control packet from the
        // remote system, and bfd.SessionState is Init or Up, the session has
        // gone down -- the local system MUST set bfd.SessionState to Down and
        // bfd.LocalDiag to 1 (Control Detection Time Expired).
        if self.rx_watchdog.is_some() {
            let _ = self
                .rx_watchdog
                .as_ref()
                .unwrap()
                .send(WatchdogReset::ReceivedPacket)
                .await;
        }
        self.state.control_packets_rx.fetch_add(1, ORDERING);
        let received_state = match p.flags.state() {
            Err(_) => {
                warn!("Invalid state, discarding");
                return;
            }
            Ok(v) => v,
        };
        // If the version number is not correct (1), the packet MUST be discarded.
        if p.flags.vers() != 1 {
            warn!("Invalid version {}, discarding", p.flags.vers());
            return;
        }
        //   If the Length field is less than the minimum correct value (24 if the A bit is clear, or 26 if the A bit is
        //   set), the packet MUST be discarded.
        if p.flags.length() < 24 || (p.flags.length() < 26 && p.flags.auth_present()) {
            warn!("Invalid packet length {}, discarding", p.flags.length());
            return;
        }
        // TODO: If the Length field is greater than the payload of the encapsulating protocol, the packet MUST be
        //   discarded.

        // If the Detect Mult field is zero, the packet MUST be discarded.
        if p.flags.detect_mult() == 0 {
            warn!("Invalid detect mult {}, discarding", p.flags.detect_mult());
            return;
        }

        //If the Multipoint (M) bit is nonzero, the packet MUST be discarded.
        if p.flags.multipoint() {
            warn!("Invalid multipoint enabled, discarding");
            return;
        }

        // If the My Discriminator field is zero, the packet MUST be discarded.
        if p.my_disc == BfdDiscriminator(0) {
            warn!("Invalid my discriminator {:?}, discarding", p.my_disc);
            return;
        }

        // If the Your Discriminator field is nonzero, it MUST be used to select the session with which this BFD packet
        //   is associated.  If no session is found, the packet MUST be discarded.
        if p.your_disc != BfdDiscriminator(0) && p.your_disc != self.state.local_discr() {
            warn!(
                "Received unexpected discriminator {:?}, discarding",
                p.your_disc
            );
            return;
        }

        // If the Your Discriminator field is zero and the State field is not Down or AdminDown, the packet MUST be
        //   discarded.
        if p.your_disc == BfdDiscriminator(0)
            && (received_state != BfdState::Down && received_state != BfdState::AdminDown)
        {
            warn!(
                "Got packet with zero discriminator and invalid state {:?}, discarding",
                received_state
            );
            return;
        }

        // If the A bit is set and no authentication is in use (bfd.AuthType is zero), the packet MUST be discarded.
        if p.flags.auth_present() && self.state.auth_type() == BfdAuthType::None {
            warn!("Got packet with auth enabled when we disagree, discarding");
            return;
        }

        // If the A bit is clear and authentication is in use (bfd.AuthType is nonzero), the packet MUST be discarded.
        if !p.flags.auth_present() && self.state.auth_type() != BfdAuthType::None {
            warn!("Got packet without auth when we expect it, discarding");
            return;
        }

        // If the A bit is set, the packet MUST be authenticated under the rules of section 6.7, based on the
        //   authentication type in use (bfd.AuthType).  This may cause the packet to be discarded.
        if p.flags.auth_present() {
            unimplemented!("Authentication is not implemented");
        }

        // Set bfd.RemoteDiscr to the value of My Discriminator.
        self.state.set_remote_discr(p.my_disc);
        // Set bfd.RemoteState to the value of the State (Sta) field.
        self.state.set_remote_session_state(received_state);
        // Set bfd.RemoteDemandMode to the value of the Demand (D) bit.
        self.state.set_remote_demand_mode(p.flags.demand());
        // Set bfd.RemoteMinRxInterval to the value of Required Min RX Interval.
        self.state.set_remote_min_rx_interval(p.required_min_rx);
        // Not in spec: store the received diagnostic
        self.state.set_last_remote_diag(p.flags.diag().unwrap());

        // If the Required Min Echo RX Interval field is zero, the transmission of Echo packets, if any, MUST cease.
        if p.required_min_echo_rx == BfdInterval::from_secs(0) {
            // TODO: implement echo thread
        }
        // If a Poll Sequence is being transmitted by the local system and the Final (F) bit in the received packet is
        //   set, the Poll Sequence MUST be terminated.
        //
        // TODO: poll stuff

        // Update the transmit interval as described in section 6.8.2.
        self.update_transmit_interval().await;

        // Update the Detection Time as described in section 6.8.4.
        self.update_detection_time(&p).await;

        // If bfd.SessionState is AdminDown
        //   Discard the packet
        if self.state.session_state() == BfdState::AdminDown {
            return;
        }

        // If received state is AdminDown
        if received_state == BfdState::AdminDown {
            // If bfd.SessionState is not Down
            if self.state.session_state() != BfdState::Down {
                // Set bfd.LocalDiag to 3 (Neighbor signaled session down)
                self.state.set_local_diag(BfdDiagnostic::NeighborDown);
                // Set bfd.SessionState to Down
                self.transition(BfdState::Down).await;
            }
        } else {
            // If bfd.SessionState ...
            match self.state.session_state() {
                BfdState::Down => {
                    // If received State is Down
                    if received_state == BfdState::Down {
                        // Set bfd.SessionState to Init
                        self.transition(BfdState::Init).await;
                    // Else if received State is Init
                    } else if received_state == BfdState::Init {
                        // Set bfd.SessionState to Up
                        self.transition(BfdState::Up).await;
                    }
                }
                BfdState::Init => {
                    // If received State is Init or Up
                    if received_state == BfdState::Init || received_state == BfdState::Up {
                        // Set bfd.SessionState to Up
                        self.transition(BfdState::Up).await;
                    }
                }
                BfdState::Up => {
                    // If received State is Down
                    if received_state == BfdState::Down {
                        // Set bfd.LocalDiag to 3 (Neighbor signaled session down)
                        self.state.set_local_diag(BfdDiagnostic::NeighborDown);
                        // Set bfd.SessionState to Down
                        self.transition(BfdState::Down).await;
                    }
                }
                BfdState::AdminDown => unreachable!("unexpected AdminDown"), // AdminDown is discarded earlier
            }

            // Check to see if Demand mode should become active or not (see section 6.6).
            // - we don't use demand mode

            // If bfd.RemoteDemandMode is 1, bfd.SessionState is Up, and bfd.RemoteSessionState is Up, Demand mode is
            // active on the remote system and
            if self.state.remote_demand_mode()
                && self.state.session_state() == BfdState::Up
                && self.state.remote_session_state() == BfdState::Up
            {
                // the local system MUST cease the periodic transmission of BFD Control packets (see section 6.8.7).
                self.periodic
                    .tx
                    .send(PeriodicControlCommand::Stop)
                    .await
                    .unwrap();
            }

            // If the Poll (P) bit is set, send a BFD Control packet to the remote system with the Poll (P) bit clear,
            // and the Final (F) bit set (see section 6.8.7).
            if p.flags.poll() {}
            // If the packet was not discarded, it has been received for purposes of the Detection Time expiration rules
            // in section 6.8.4.
        }
    }

    // https://datatracker.ietf.org/doc/html/rfc5880#section-6.8.2
    async fn update_transmit_interval(&mut self) {
        let new_base_interval = std::cmp::max(
            self.state.desired_min_tx_interval(),
            self.state.remote_min_rx_interval(),
        );
        if new_base_interval != self.cur_interval {
            self.periodic
                .tx
                .send(PeriodicControlCommand::SetMinInterval(new_base_interval))
                .await
                .unwrap()
        }
        self.cur_interval = new_base_interval;
    }

    // https://datatracker.ietf.org/doc/html/rfc5880#section-6.8.4
    async fn update_detection_time(&mut self, p: &BfdPacket) {
        let old = self.state.detection_time();
        self.state.set_detection_time(if !self.state.demand_mode() {
            (p.flags.detect_mult() as u32
                * std::cmp::max(self.state.required_min_rx_interval(), p.desired_min_tx))
            .into()
        } else {
            (self.state.detect_mult() as u32
                * std::cmp::max(
                    self.state.desired_min_tx_interval(),
                    self.state.remote_min_rx_interval(),
                ))
            .into()
        });
        if old != self.state.detection_time() {
            self.start_watchdog().await
        }
    }
}

#[derive(Debug)]
enum EventDispatcherCommand {
    AddSink(Box<dyn EventMessageSink>),
    DispatchMessage(EventMessage),
}

struct EventDispatcher {
    event_sinks: Vec<mpsc::Sender<EventMessage>>,
    event_sink_handles: Vec<JoinHandle<()>>,
    tx: mpsc::Sender<EventDispatcherCommand>,
    rx: mpsc::Receiver<EventDispatcherCommand>, // TODO: there must be a better way to do this
}

impl EventDispatcher {
    fn new(
        channel_tx: mpsc::Sender<EventDispatcherCommand>,
        channel_rx: mpsc::Receiver<EventDispatcherCommand>,
    ) -> Self {
        let event_sinks = Vec::new();
        let event_sink_handles = Vec::new();

        Self {
            event_sinks,
            event_sink_handles,
            tx: channel_tx,
            rx: channel_rx,
        }
    }
    fn add_sink(&mut self, sink: Box<dyn EventMessageSink>) {
        // Save the transmit channel
        self.event_sinks.push(sink.channel());
        // Run the sink
        self.event_sink_handles.push(sink.run());
    }
    async fn dispatch(&self, event: EventMessage) {
        debug!(
            "EventDispatcher dispatching event {:?} to sinks {:?}",
            event, self.event_sinks
        );
        for sink in &self.event_sinks {
            sink.send(event.clone()).await.unwrap();
        }
    }
    async fn run(mut self) {
        info!("Event dispatcher running");
        while let Some(cmd) = self.rx.recv().await {
            debug!("EventDispatcher got command {:?}", cmd);
            match cmd {
                EventDispatcherCommand::AddSink(sink) => self.add_sink(sink),
                EventDispatcherCommand::DispatchMessage(event) => self.dispatch(event).await,
            }
        }
    }
}

#[derive(Clone)]
struct EventDispatcherHandle {
    channel_tx: mpsc::Sender<EventDispatcherCommand>,
}

impl EventDispatcherHandle {
    async fn new() -> Self {
        let (channel_tx, channel_rx) = mpsc::channel(16);
        let dispatcher = EventDispatcher::new(channel_tx.clone(), channel_rx);
        tokio::spawn(async move { dispatcher.run().await });
        Self { channel_tx }
    }
    async fn dispatch(&self, event: EventMessage) {
        self.channel_tx
            .send(EventDispatcherCommand::DispatchMessage(event))
            .await
            .unwrap();
    }
    async fn add_sink(&self, sink: Box<dyn EventMessageSink>) {
        self.channel_tx
            .send(EventDispatcherCommand::AddSink(sink))
            .await
            .unwrap()
    }
}

struct Bfdd {
    local: SocketAddr,
    peers: Vec<SocketAddr>,
    sessions_by_ip: HashMap<(IpAddr, IpAddr), BfdDiscriminator>,
    sessions_by_discr: HashMap<BfdDiscriminator, BfdSessionHandle>,

    control_socket: Arc<UdpSocket>,

    event_dispatcher: EventDispatcherHandle,
}

impl Bfdd {
    async fn new() -> Result<Self, Box<dyn Error>> {
        // let local = SocketAddr::from_str("192.168.255.2:3784").unwrap();
        let local = SocketAddr::new(Ipv6Addr::from_str("2001:db8::2").unwrap().into(), 3784);
        let peers = vec![
            // SocketAddr::from_str("192.168.122.132:3784").unwrap(),
            // SocketAddr::from_str("192.168.255.1:3784").unwrap(),
            SocketAddr::new(Ipv6Addr::from_str("2001:db8::1").unwrap().into(), 3784),
        ];
        let sessions_by_ip = HashMap::new();
        let sessions_by_discr = HashMap::new();

        // TODO: Bind only the sockets we need for the configuration, including multihop and run them in different tasks.
        info!("Binding on {:?}", local);
        let control_socket = Arc::new(UdpSocket::bind(local).await?);
        // If BFD authentication is not in use on a session, all BFD Control packets for the session MUST be sent with a
        // Time to Live (TTL) or Hop Limit value of 255.
        set_ttl_or_hop_limit(&control_socket, 255)?;
        let echo_socket = Arc::new(UdpSocket::bind(SocketAddr::new(local.ip(), ECHO_PORT)).await?);

        let event_dispatcher = EventDispatcherHandle::new().await;
        event_dispatcher
            .add_sink(Box::new(ScriptHookSink::new()))
            .await;

        Ok(Self {
            local,
            peers,
            sessions_by_ip,
            sessions_by_discr,
            control_socket,
            event_dispatcher,
        })
    }
    async fn start_peers(&mut self) {
        let mut rng = rand::thread_rng();
        for peer in &self.peers {
            let mut local_discr = BfdDiscriminator(0);
            while local_discr.0 == 0 || self.sessions_by_discr.get(&local_discr).is_some() {
                local_discr = BfdDiscriminator(rng.gen());
            }
            let session = BfdSessionHandle::new(
                self.local.ip(),
                peer.ip(),
                local_discr,
                self.event_dispatcher.clone(),
            )
            .await;

            self.sessions_by_ip
                .insert((self.local.ip(), peer.ip()), local_discr);
            self.sessions_by_discr.insert(local_discr, session);
        }
    }
    fn session_for_packet(
        &self,
        addr: &SocketAddr,
        packet: &BfdPacket,
    ) -> Option<&BfdSessionHandle> {
        // Try to find a configured session for the remote IP if it doesn't send our discriminator, otherwise
        // use the discriminator that was sent.

        // https://datatracker.ietf.org/doc/html/rfc5880#section-6.3
        //
        // Once the remote end echoes back the local discriminator, all further received packets are
        // demultiplexed based on the Your Discriminator field only (which means that, among other things, the
        // source address field can change or the interface over which the packets are received can change, but
        // the packets will still be associated with the proper session).
        let our_disc = if packet.your_disc == BfdDiscriminator(0) {
            if let Some(discr) = self
                .sessions_by_ip
                .get(&(self.control_socket.local_addr().unwrap().ip(), addr.ip()))
            {
                // TODO: Probably we should destroy the old session here?
                debug!(
                    "Found session for unknown discriminator from {}: {}",
                    addr.ip(),
                    discr
                );
                discr
            } else {
                warn!("Unable to match packet to session with {}", addr.ip());
                return None;
            }
        } else {
            &packet.your_disc
        };
        self.sessions_by_discr.get(our_disc)
    }
    async fn rx_packet(&self, addr: SocketAddr, packet: BfdPacket) {
        if let Some(session) = self.session_for_packet(&addr, &packet) {
            session
                .tx
                .send(SessionControlCommand::RxPacket(packet))
                .await
                .unwrap();
        } else {
            warn!("Unable to match packet to session with");
            return;
        }
    }
    async fn run(mut self) -> io::Result<()> {
        info!("rust-bfd starting up");
        self.start_peers().await;
        info!(
            "configured sessions: {}",
            future::join_all(
                self.sessions_by_discr
                    .values()
                    .map(|session| async { session.get_stats().await.unwrap().to_string() })
            )
            .await
            .join("\n\t")
        );

        let rx_task =
            task::spawn(async move {
                let mut buf = [0; 1024];
                loop {
                    // TODO: All received BFD Control packets that are demultiplexed to the session MUST be discarded if
                    // the received TTL or Hop Limit is not equal to 255. Need raw sockets for this.
                    //
                    // TODO: Maybe it is okay to receive in the session-specific socket instead of this global one? But
                    // the RFC mentions that we must demux based on discriminator only.
                    let (len, addr) = self.control_socket.recv_from(&mut buf).await.unwrap(); // TODO: fallibility?
                    if let Ok((_leftover, packet)) = BfdPacket::parse(&buf[..len]) {
                        debug!("rx packet ({:?}): {:?}", addr.ip(), packet);
                        self.rx_packet(addr, packet).await
                    } else {
                        warn!("Failed to parse packet");
                    }

                    info!(
                        "configured sessions: {}",
                        future::join_all(self.sessions_by_discr.values().map(|session| async {
                            session.get_stats().await.unwrap().to_string()
                        }))
                        .await
                        .join("\n\t")
                    );
                }
            });

        tokio::join!(rx_task)
            .0
            .expect("Unable to join on the receive thread");

        Ok(())
    }
}

#[tokio::main]
async fn main() -> io::Result<()> {
    env_logger::Builder::from_env(Env::default().default_filter_or("info"))
        .format_timestamp_micros()
        .init();

    match Bfdd::new().await {
        Ok(bfdd) => bfdd.run().await,
        Err(e) => Err(std::io::Error::new(ErrorKind::Other, e.to_string())),
    }
}
