use core::{fmt, str};
use std::{
    ffi::{OsStr, OsString},
    net::IpAddr,
    path::PathBuf,
    process::ExitStatus,
};

use itertools::Itertools;
use log::{debug, info, warn};
use nix::unistd::{self, eaccess};
use tokio::{
    fs, process,
    sync::mpsc,
    task::{self, JoinHandle},
};

use crate::{BfdDiscriminator, BfdState};

#[derive(Debug, Clone)]
pub enum EventMessage {
    StateChange(StateChangeEvent),
}

#[derive(Debug, Clone)]
pub struct StateChangeEvent {
    pub local_discr: BfdDiscriminator,
    pub remote_discr: BfdDiscriminator,
    pub local_ip: IpAddr,
    pub remote_ip: IpAddr,
    pub from_state: BfdState,
    pub to_state: BfdState,
}

pub trait EventMessageSink: Send + Sync + fmt::Debug {
    fn channel(&self) -> mpsc::Sender<EventMessage>;
    fn run(self: Box<Self>) -> JoinHandle<()>;
}

#[derive(Debug)]
pub struct ScriptHookSink {
    script_dir: PathBuf,
    tx_channel: mpsc::Sender<EventMessage>,
    rx_channel: mpsc::Receiver<EventMessage>,
}

impl EventMessageSink for ScriptHookSink {
    fn channel(&self) -> mpsc::Sender<EventMessage> {
        self.tx_channel.clone()
    }
    fn run(self: Box<Self>) -> JoinHandle<()> {
        task::spawn(async move { self.event_loop().await })
    }
}

impl ScriptHookSink {
    pub fn new() -> Self {
        let (tx_channel, rx_channel) = mpsc::channel(16);
        Self {
            script_dir: "/etc/rust-bfd/hooks.d/".into(),
            tx_channel,
            rx_channel,
        }
    }
    async fn script_runner<S: AsRef<OsStr>>(script_path: S, args: &[S], environment: &[(S, S)]) {
        let mut cmd = process::Command::new("sh");
        cmd.env_clear()
            .arg0(&script_path)
            .arg(&script_path)
            .args(args)
            .envs(environment.iter().map(|(k, v)| (k, v)));
        debug!("Running command {:?}", cmd);
        let result = cmd.output().await;
        match result {
            Ok(output) => {
                info!(
                    "Script `{} {}` completed with status {}",
                    cmd.as_std().get_program().to_string_lossy(),
                    cmd.as_std()
                        .get_args()
                        .map(|arg| arg.to_string_lossy())
                        .join(" "),
                    output.status
                );
                if !output.status.success() {
                    if output.stdout.len() > 0 {
                        warn!(
                            " stdout: \n{}",
                            str::from_utf8(output.stdout.as_slice()).unwrap()
                        );
                    }
                    if output.stderr.len() > 0 {
                        warn!(
                            " stderr: \n{}",
                            str::from_utf8(output.stderr.as_slice()).unwrap()
                        );
                    }
                }
            }
            Err(e) => {
                warn!("Failed to execute {:?}, {}", cmd, e);
            }
        }
    }
    async fn send_state_change(&self, event: &StateChangeEvent) {
        debug!("Got a state change event! {:?}, running scripts", event);
        let mut files = match fs::read_dir(&self.script_dir).await {
            Ok(files) => files,
            Err(e) => {
                warn!(
                    "Error opening script directory `{}` ({}), no hooks executed.",
                    self.script_dir.as_os_str().to_string_lossy(),
                    e
                );
                return;
            }
        };
        let mut scripts = Vec::new();
        while let Ok(Some(direntry)) = files.next_entry().await {
            // Check if entry is a file
            if direntry.file_type().await.is_ok_and(|ft| ft.is_file()) {
                // Check if it's executable
                let path = direntry.path();
                if unistd::eaccess(&path, unistd::AccessFlags::X_OK).is_ok() {
                    // Add to scripts array
                    scripts.push(path);
                }
            }
        }
        // Execute in lexicographic order, so sort the list of found scripts
        scripts.sort_unstable_by_key(|path| {
            path.file_name()
                .map_or(OsString::new(), |file| file.to_owned())
        });

        if scripts.len() > 0 {
            // Build context
            let mut args = Vec::<&OsStr>::new();
            let mut env = Vec::<(&OsStr, &OsStr)>::new();

            let remote_ip_str = OsString::from(event.remote_ip.to_string());
            let local_ip_str = OsString::from(event.local_ip.to_string());
            let remote_discr_str = OsString::from(event.remote_discr.to_string());
            let local_discr_str = OsString::from(event.local_discr.to_string());

            args.push(event.to_state.into());
            args.push(&remote_ip_str);
            env.push((OsStr::new("BFD_LAST_STATE"), event.from_state.into()));
            env.push((OsStr::new("BFD_STATE"), event.to_state.into()));
            env.push((OsStr::new("BFD_LOCAL_ADDR"), &local_ip_str));
            env.push((OsStr::new("BFD_PEER_ADDR"), &remote_ip_str));
            env.push((OsStr::new("BFD_LOCAL_DISCRIMINATOR"), &local_discr_str));
            env.push((OsStr::new("BFD_PEER_DISCRIMINATOR"), &remote_discr_str));
            for script in scripts {
                Self::script_runner(script.as_os_str(), &args, &env).await
            }
        }
    }
    pub async fn handle_event(&self, event: &EventMessage) {
        match event {
            EventMessage::StateChange(change) => self.send_state_change(change).await,
        }
    }
    async fn event_loop(mut self) {
        info!("ScriptHookSink started");
        while let Some(event) = self.rx_channel.recv().await {
            debug!("ScriptHookSink got event {:?}", event);
            self.handle_event(&event).await
        }
    }
}
